import os
import json

from jsonschema import validate

DEFAULT_SCHEMA = os.path.join(os.path.dirname(__file__), "config_schema.json")


def validate_config(data, schema=None):
    """validate the data according to the schema"""
    if schema is None:
        with open(DEFAULT_SCHEMA) as f:
            schema = json.load(f)
    validate(data, schema)



"""This module abstract the 'modules' content type"""


class Module(object):
    """Define a single module"""
    def __init__(self, name, stream, profiles=None):
        self.name = name
        self.stream = str(stream)
        self.profiles = profiles

    def __repr__(self):
        return '<Module: %s>' % self.name


class Modules(object):
    """group of modules"""

    def __init__(self, include):
        """
        Args:
            include(list): a list of Module instances
        """
        self.whitelist = include

    def __getitem__(self, index):
        return self.whitelist[index]

    @classmethod
    def load_from_dict(cls, data):
        """create instances of moudles from a dictionary"""
        include = []

        for item in data.get('include', []):
            module = Module(**item)
            include.append(module)

        return cls(include)



import os
import logging

from six.moves.urllib.parse import urlparse

from ._impl.loaders import LocalLoader, GitlabLoader


DEFAULT_UBI_REPO = os.getenv("DEFAULT_UBI_REPO", "")

LOG = logging.getLogger('ubiconfig')


class LoaderError(RuntimeError):
    pass


def get_loader(source=None):
    """Get a Loader instance which is used to load configurations.

    ``source`` should be provided as one of the following:

        URL
            A URL of a remote git repo containing UBI config files.
            Currently, only Gitlab is supported.

        local path
            A path to a local directory containing UBI config files.

        :any:`None`
            If none/omitted, the value of the ``DEFAULT_UBI_REPO``
            environment variable is used. If this is unset, an
            exception is raised.

    After the loader is constructed, it can be used to load config files
    when given relative paths to config files.

    .. code-block:: python

        # use default config source
        >>> loader = get_loader()
        >>> config_ubi7 = loader.load('ubi7.yaml')
        >>> config_ubi7.content_sets.rpm.input
        # loader can be used repeatedly
        >>> config_ubi8 = loader.load('ubi8.yaml')

        # or use a local directory
        >>> loader = get_loader('/my/config/dir)
        >>> config = loader.load('path/to/configfile.yaml')
    """
    if not source:
        source = DEFAULT_UBI_REPO

    if not source:
        msg = 'Please either set a source or define DEFAULT_UBI_REPO in your environment'
        raise LoaderError(msg)

    parsed = urlparse(source)
    if parsed.netloc:
        # It's a URL, use the gitlab loader
        return GitlabLoader(source)

    # It should be a local path
    if not os.path.isdir(source):
        raise LoaderError("'%s' is not an existing directory" % source)

    return LocalLoader(source)



class Loader(object):
    """Load UBI configuration.

    Don't construct instances of this class directly;
    use the :func:`~ubiconfig.get_loader` function.
    """

    # This is a base class only. See implementations alongside this one
    # under the loaders module.

    def load(self, file_name):
        """
        Load a single configuration file and return a :class:`UbiConfig` object.

        The given file_name should be a relative path to a YAML file
        in the loader's config source (e.g. relative path to file within a
        git repo or local directory).
        """
        raise NotImplementedError()

    def load_all(self, recursive=False):
        """Get the list of config files from repo and call load on every file.
        Return a list of :class:`UbiConfig` objects.

        If recursive is set, it will walk through the submodules, no matter local
        or remote
        """
        raise NotImplementedError()



import logging
import os

import yaml
from jsonschema.exceptions import ValidationError

from ubiconfig.utils.config_validation import validate_config
from ubiconfig.config_types import UbiConfig


LOG = logging.getLogger('ubiconfig')


class LocalLoader(object):
    """Load configuration from a local directory tree."""
    def __init__(self, path):
        self._path = path

    def load(self, file_name):
        file_path = os.path.join(self._path, file_name)
        LOG.info("Loading configuration file locally: %s", file_path)

        with open(file_path, 'r') as f:
            config_dict = yaml.safe_load(f)
        # validate input data
        validate_config(config_dict)

        return UbiConfig.load_from_dict(config_dict, file_name)

    def load_all(self, recursive=False):
        ubi_configs = []

        file_list = self._get_local_file_list(recursive)
        for file in file_list:
            LOG.debug("Now loading %s", file)
            try:
                ubi_configs.append(self.load(file))
            except yaml.YAMLError:
                LOG.error("%s FAILED loading because of Syntax error, Skip for now", file)
                continue
            except ValidationError as e:
                LOG.error("%s FAILED schema validation:\n%s\nSkip for now", file, e)
                continue

        return ubi_configs

    def _get_local_file_list(self, recursive):
        """
        Get the config file list from local. If recusive is set, then it would walk
        through the sub-modules.
        """
        LOG.info('Getting the local config file list')
        file_list = []
        if recursive:
            for root, _, files in os.walk(self._path):
                files = [os.path.join(root, f) for f in files if f.endswith(('.yaml', '.yml'))]
                file_list.extend(files)
        else:
            file_list = [file for file in os.listdir(self._path)
                         if file.endswith(('yaml', '.yml'))]

        return file_list



import logging
import yaml
import requests

from jsonschema.exceptions import ValidationError

from ubiconfig.utils.api.gitlab import RepoApi
from ubiconfig.utils.config_validation import validate_config
from ubiconfig.config_types import UbiConfig

from .base import Loader

LOG = logging.getLogger('ubiconfig')


class GitlabLoader(Loader):
    """Load configuration from a remote repo on gitlab."""
    def __init__(self, url):
        self._url = url
        self._session = requests.Session()
        self._repo_api = RepoApi(self._url.rstrip('/'))
        self._files_branch_map = self._pre_load()

    def load(self, file_name):
        # find the right branch from the mapping
        branch = self._files_branch_map[file_name]

        config_file_url = self._repo_api.get_file_content_api(file_name, branch)
        LOG.info("Loading configuration file from remote: %s", file_name)
        response = self._session.get(config_file_url)
        response.raise_for_status()

        config_dict = yaml.safe_load(response.content)
        # validate input data
        validate_config(config_dict)

        return UbiConfig.load_from_dict(config_dict, file_name)

    def load_all(self, recursive=False):
        ubi_configs = []
        for file in self._files_branch_map:
            LOG.debug("Now loading %s from branch %s", file, self._files_branch_map[file])
            try:
                ubi_configs.append(self.load(file))
            except yaml.YAMLError:
                LOG.error("%s FAILED loading because of Syntax error, Skip for now", file)
                continue
            except ValidationError as e:
                LOG.error("%s FAILED schema validation:\n%s\nSkip for now", file, e)
                continue

        return ubi_configs

    def _pre_load(self):
        """Iterate all branches to get a mapping of {file_path: branch,...}
        """
        files_branch_map = {}
        branches = self._get_branches()
        LOG.info("Loading config files from all branches: %s", branches)
        for branch in branches:
            page = 1
            while True:
                file_list_api = self._repo_api.get_file_list_api(branch=branch,
                                                                 page=page)
                response = self._session.get(file_list_api)
                file_list = [file['path'] for file in response.json()
                             if file['name'].endswith(('.yaml', '.yml'))]
                for file in file_list:
                    files_branch_map[file] = branch
                if page >= int(response.headers.get('X-Total-Pages', 1)):
                    break
                page += 1
        return files_branch_map

    def _get_branches(self):
        """Get all the branches of a given repo"""
        LOG.info("Getting branches of the repo")
        branches_list_api = self._repo_api.get_branch_list_api()
        json_response = self._session.get(branches_list_api).json()
        if not json_response:
            raise RuntimeError('Please check %s is in right format' % self._url)
        branches = [b['commit']['id'] for b in json_response]
        return branches



from .modules import Modules
from .packages import Packages
from .content_sets import ContentSetsMapping


class UbiConfig(object):
    """Wrap all UBI related configurations
    Examples to access different configurations:
    Modules:
      config.modules[0].whitelist[0].name
    Packages:
      config.packages.whitelist[0].name
      config.packages.blacklist[0].name
    ContentSets:
      config.content_sets.rpm.input
      config.content_sets.debuginfo.output"""
    def __init__(self, cs, pkgs, mds, file_name):
        self.content_sets = cs
        self.packages = pkgs
        self.modules = mds
        self.file_name = file_name

    def __repr__(self):
        return self.file_name

    @classmethod
    def load_from_dict(cls, data, file_name):
        m_data = Modules.load_from_dict(data.get('modules', {}))
        pkgs = data.get('packages', {})
        pkgs_data = Packages(pkgs.get('include', []),
                             pkgs.get('exclude', []),
                             data.get('arches', []))
        cs_map = ContentSetsMapping.load_from_dict(data['content_sets'])
        # use the simplified file name
        file_name = file_name.split('/')[-1]

        return cls(cs=cs_map, pkgs=pkgs_data, mds=m_data, file_name=file_name)



"""This module abstracts the 'content_sets' mappings"""


class ContentSetMapping(object):

    def __init__(self, input_content, output_content):
        self.input = input_content
        self.output = output_content


class Rpm(ContentSetMapping):

    @property
    def type(self):
        return 'rpm'


class Srpm(ContentSetMapping):

    @property
    def type(self):
        return 'srpm'


class Debuginfo(ContentSetMapping):

    @property
    def type(self):
        return 'debuginfo'


class ContentSetsMapping(object):
    def __init__(self, rpm, srpm, debuginfo):
        self.rpm = rpm
        self.srpm = srpm
        self.debuginfo = debuginfo

    @classmethod
    def load_from_dict(cls, data):
        rpm_data = data.get('rpm', {})
        rpm = Rpm(rpm_data.get('input', ''), rpm_data.get('output', ''))

        srpm_data = data.get('srpm', {})
        srpm = Srpm(srpm_data.get('input', ''), srpm_data.get('output', ''))

        debug_data = data.get('debuginfo', {})
        debuginfo = Debuginfo(debug_data.get('input', ''), debug_data.get('output', ''))

        return cls(rpm, srpm, debuginfo)

    def export_dict(self):
        """Return a dictionary such as {type: (input, outpt)}"""
        attributes = [self.rpm, self.srpm, self.debuginfo]
        return dict((m.type, (m.input, m.output)) for m in attributes)



from setuptools import setup, find_packages


def get_description():
    return 'A Python Library for accessing Universal Base Image configuration'


def get_long_description():
    with open('README.md') as f:
        text = f.read()

    # Long description is everything after README's initial heading
    idx = text.find('\n\n')
    return text[idx:]


def get_requirements():
    with open('requirements.txt') as f:
        return f.read().splitlines()


setup(
    name='ubi-config',
    version='1.0.1',
    author='',
    author_email='',
    packages=find_packages(exclude=['tests', 'tests.*']),
    package_data={'ubiconfig': ['utils/config_schema.json']},
    url='https://github.com/release-engineering/ubi-config',
    license='GNU General Public License',
    description=get_description(),
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    install_requires=get_requirements(),
    project_urls={
        'Documentation': 'https://release-engineering.github.io/ubi-config/',
    },
)
